from pickle import loads, dump, dumps

def encode(object: any):
    if type(object) == str:
        return dumps(object)
    else:
        return dumps(object)

def decode(data: bytes):
    return loads(data)


"""
The server's view of what is a client; not the main client.
"""

from pydatanet.data import encode, decode

class Client:
    def __init__(self, ip: str, port: int, conn):
        self.ip = ip
        self.port = port
        self.conn = conn
    
    def close(self):
        return self.conn.close()

    def getIp(self):
        return self.ip

    def getPort(self):
        return self.port
    
    def _send(self, data: bytes):
        print(data)
        self.conn.send(data)

    def send(self, data: any):
        self._send(encode(data))


import enum
import socket
import threading
from socket import AF_INET, SO_REUSEADDR, SOCK_STREAM, SOCK_DGRAM, SOL_SOCKET

from pydatanet.server.client import Client
from pydatanet.data import encode, decode

class ServerType(enum.Enum):
    TCP = 0
    UDP = 1

class ServerStatus(enum.Enum):
    NOINIT = 0
    INIT = 1

BUFFER_SIZE = 2048

class Server:
    def __init__(self, host: str, port: int, serverType = ServerType.TCP, maxConn: int = 256):
        self.host = host
        self.port = port
        self.socket = None
        self.maxConn = maxConn
        self.status = ServerStatus.NOINIT
        self.clients = []
        self.serverType = serverType

        self.dataReceiveListeners = []
        self.onJoinListeners = []

        if serverType == ServerType.TCP:
            print("* creating TCP socket")
            self.socket = socket.socket(AF_INET, SOCK_STREAM)
        elif serverType == ServerType.UDP:
            print("* creating UDP socket")
            self.socket = socket.socket(AF_INET, SOCK_DGRAM)
        else:
            raise ValueError(f"Invalid server type '{serverType}'!")

        self.socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)

    def on_recv(self, sender, data):
        """
        Triggered when someone sends a packet. Do not explicitly trigger as unexpected behaviour may occur.
        """
        data = decode(data)
        for listenFunc in self.dataReceiveListeners:
            listenFunc(sender, data)

    def on_recv_bind(self, func):
        """
        Bind a function to when someone sends data.
        """
        self.dataReceiveListeners.append(func)

    def connect(self):
        """
        Bind to that socket and start listening for connections.
        """
        if self.socket is not None:
            self.socket.bind((self.host, self.port))
            
            if self.serverType == ServerType.TCP:
                # TCP shenanigans that UDP does not fancy.
                self.socket.listen(self.maxConn)
                print(f"* socket is now listening at {self.host}:{self.port}")

            self.status = ServerStatus.INIT
            self.heartbeat()
            print("* server is now in heartbeating mode.")

    def _heartbeat(self):
        """
        Server heartbeat/poll method.
        """
        while self.status == ServerStatus.INIT:
            if self.serverType == ServerType.UDP:
                self.udp_poll()
            else:
                self.tcp_poll()

    def heartbeat(self):
        """
        Call this function once to start server polling. Automatically called by [Server.connect]
        """
        thr = threading.Thread(target = self._heartbeat, args=())
        thr.start()

    def getFromIp(self, ip: str, port: int):
        for client in self.clients:
            if client.getIp() == ip and client.getPort() == port:
                return client

        return None

    def on_join_hook(self, func):
        self.onJoinListeners.append(func)

    def tcp_poll(self):
        """
        TCP socket polling method.
        """
        # Receive data from existing connections.
        for client in self.clients:
            data = client.conn.recv(BUFFER_SIZE)
            if not data:
                print(f"\t\t* [{client.ip}:{client.port}] did not send any data in return, indicating a disconnect.")
                client.close()
                self.clients.remove(client)
            
            self.on_recv(client, data)

        # Look out for new connections.
        conn, addr = self.socket.accept()
        client = Client(addr[0], addr[1], conn)
        print(f"\t* new connection has been established [{addr[0]}:{addr[1]}]")

        for func in self.onJoinListeners:
            func(client)

        self.clients.append(
            client
        )

    def udp_poll(self):
        data, sender = self.socket.recvfrom(BUFFER_SIZE)

        self.on_recv(Client(sender[0], sender[1], None), data)
    
    def stop(self):
        """
        Stop the server by closing the socket.
        """
        print(f"* socket at {self.host}:{self.port} is now closing.")
        self.socket.close()

    def __del__(self):
        self.stop()


"""
The actual client that connects/sends data. Not the server representation.
"""

import socket
import enum
from socket import AF_INET, SO_REUSEADDR, SOCK_STREAM, SOCK_DGRAM, SOL_SOCKET
import threading

from pydatanet.data import encode, decode

BUFFER_SIZE = 2048

class ClientType(enum.Enum):
    TCP = 0
    UDP = 1

class ConnectStatus(enum.Enum):
    CONNECTED = 0
    DISCONNECTED = 1

class Client:
    def __init__(self, clientType = ClientType.TCP):
        self.clientType = clientType
        self.socket = None
        self.onRecvTCP = []
        self.onRecvUDP = []
        self.status = ConnectStatus.DISCONNECTED
        self.addr = ""
        self.port = 0

        if clientType == ClientType.TCP:
            self.socket = socket.socket(AF_INET, SOCK_STREAM)
        elif clientType == ClientType.UDP:
            self.socket = socket.socket(AF_INET, SOCK_DGRAM)
        else:
            raise ValueError(f"Invalid client type '{clientType}'!")

    def connect(self, address: str, port: int, autoPoll: bool = True):
        print(f"* connecting to [{address}:{port}]")
        if self.clientType == ClientType.TCP:
            self.socket.connect(
                (address, port)
            )

        self.addr = address
        self.port = port

        self.status = ConnectStatus.CONNECTED
        if autoPoll: self.heartbeat()
        print(f"\t* connected to [{address}:{port}]")

    def send(self, data: any):
        if self.clientType == ClientType.TCP:
            self.socket.send(encode(data))
        elif self.clientType == ClientType.UDP:
            self.socket.sendto(encode(data), (self.addr, self.port))

    def heartbeat(self):
        threading.Thread(target=self._heartbeatLoop, args=()).start()

    def _heartbeatLoop(self):
        while True:
            self._heartbeat()

    def _heartbeat(self):
        if self.status != ConnectStatus.CONNECTED:
            return
            
        if self.clientType == ClientType.TCP:
            self.poll_TCP()
        elif self.clientType == ClientType.UDP:
            self.poll_UDP()
        else:
            raise ValueError(self.clientType)

    def hook_tcp_recv(self, func):
        """
        Hook a function to the event in which a packet is received (TCP).
        """
        self.onRecvTCP.append(func)

    def hook_udp_recv(self, func):
        """
        Hook a function to the event in which a packet is received (TCP).
        """
        self.onRecvUDP.append(func)
    
    def poll_TCP(self):
        data = self.socket.recv(BUFFER_SIZE)
        if not data:
            print("\t* we have been disconnected from the server.")
            self.status = ConnectStatus.DISCONNECTED
            return
            
        data = decode(data)
        
        for event in self.onRecvTCP:
            event(data)

    def poll_UDP(self):
        data = self.socket.recvfrom(BUFFER_SIZE)
        data = decode(data)
        
        for event in self.onRecvUDP:
            event(data)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydatanet",
    version="0.1.1-rc9",
    author="xTrayambak",
    author_email="xtrayambak@gmail.com",
    description="A fancy networking library which supports both the protocols (UDP support WIP); and makes it hard to shoot yourself in the foot.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/xTrayambak/pysyn",
    project_urls={
        "Bug Tracker": "https://github.com/xTrayambak/pysyn/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src/"},
    packages=setuptools.find_packages(where="src/"),
    python_requires=">=3.8",
)